@@ -30,7 +30,8 @@
 #include <glib.h>
 #include <gdbus.h>
 
-#include "logging.h"
+#include "log.h"
+#include "transfer.h"
 #include "session.h"
 #include "pbap.h"
 
@@ -245,7 +246,7 @@ static gint pbap_set_path(struct session_data *session, const char *path)
 	if (!path)
 		return OBEX_RSP_BAD_REQUEST;
 
-	if (pbapdata->path != NULL && 	g_str_equal(pbapdata->path, path))
+	if (pbapdata->path != NULL && g_str_equal(pbapdata->path, path))
 		return 0;
 
 	if (gw_obex_chdir(session->obex, "", &err) == FALSE) {
@@ -280,7 +281,8 @@ fail:
 static void read_return_apparam(struct session_data *session,
 				guint16 *phone_book_size, guint8 *new_missed_calls)
 {
-	GwObexXfer *xfer = session->xfer;
+	struct transfer_data *transfer = session->pending->data;
+	GwObexXfer *xfer = transfer->xfer;
 	unsigned char *buf;
 	size_t size = 0;
 
@@ -310,7 +312,7 @@ static void read_return_apparam(struct session_data *session,
 			if (hdr->len == PHONEBOOKSIZE_LEN) {
 				guint16 val;
 				memcpy(&val, hdr->val, sizeof(val));
-				*phone_book_size = val;
+				*phone_book_size = get_be16(&val);
 			}
 			break;
 		case NEWMISSEDCALLS_TAG:
@@ -329,33 +331,44 @@ static void read_return_apparam(struct session_data *session,
 }
 
 static void pull_phonebook_callback(struct session_data *session,
-					void *user_data)
+					GError *err, void *user_data)
 {
+	struct transfer_data *transfer = session->pending->data;
 	DBusMessage *reply;
 	char *buf = "";
 
+	if (session->msg == NULL)
+		goto done;
+
 	reply = dbus_message_new_method_return(session->msg);
 
-	if (session->filled > 0)
-		buf = session->buffer;
+	if (transfer->filled > 0)
+		buf = transfer->buffer;
 
 	dbus_message_append_args(reply,
 			DBUS_TYPE_STRING, &buf,
 			DBUS_TYPE_INVALID);
 
-	session->filled = 0;
+	transfer->filled = 0;
 	g_dbus_send_message(session->conn, reply);
 	dbus_message_unref(session->msg);
 	session->msg = NULL;
+
+done:
+	transfer_unregister(transfer);
 }
 
 static void phonebook_size_callback(struct session_data *session,
-					void *user_data)
+					GError *err, void *user_data)
 {
+	struct transfer_data *transfer = session->pending->data;
 	DBusMessage *reply;
 	guint16 phone_book_size;
 	guint8 new_missed_calls;
 
+	if (session->msg == NULL)
+		goto done;
+
 	reply = dbus_message_new_method_return(session->msg);
 
 	read_return_apparam(session, &phone_book_size, &new_missed_calls);
@@ -364,30 +377,37 @@ static void phonebook_size_callback(struct session_data *session,
 			DBUS_TYPE_UINT16, &phone_book_size,
 			DBUS_TYPE_INVALID);
 
-	session->filled = 0;
+	transfer->filled = 0;
 	g_dbus_send_message(session->conn, reply);
 	dbus_message_unref(session->msg);
 	session->msg = NULL;
+
+done:
+	transfer_unregister(transfer);
 }
 
 static void pull_vcard_listing_callback(struct session_data *session,
-					void *user_data)
+					GError *err, void *user_data)
 {
+	struct transfer_data *transfer = session->pending->data;
 	GMarkupParseContext *ctxt;
 	DBusMessage *reply;
 	DBusMessageIter iter, array;
 	int i;
 
+	if (session->msg == NULL)
+		goto complete;
+
 	reply = dbus_message_new_method_return(session->msg);
 
-	if (session->filled == 0)
+	if (transfer->filled == 0)
 		goto done;
 
-	for (i = session->filled - 1; i > 0; i--) {
-		if (session->buffer[i] != '\0')
+	for (i = transfer->filled - 1; i > 0; i--) {
+		if (transfer->buffer[i] != '\0')
 			break;
 
-		session->filled--;
+		transfer->filled--;
 	}
 
 	dbus_message_iter_init_append(reply, &iter);
@@ -396,17 +416,19 @@ static void pull_vcard_listing_callback(struct session_data *session,
 			DBUS_TYPE_STRING_AS_STRING DBUS_TYPE_STRING_AS_STRING
 			DBUS_STRUCT_END_CHAR_AS_STRING, &array);
 	ctxt = g_markup_parse_context_new(&listing_parser, 0, &array, NULL);
-	g_markup_parse_context_parse(ctxt, session->buffer,
-					session->filled, NULL);
+	g_markup_parse_context_parse(ctxt, transfer->buffer,
+					transfer->filled, NULL);
 	g_markup_parse_context_free(ctxt);
 	dbus_message_iter_close_container(&iter, &array);
 
-	session->filled = 0;
+	transfer->filled = 0;
 
 done:
 	g_dbus_send_message(session->conn, reply);
 	dbus_message_unref(session->msg);
 	session->msg = NULL;
+complete:
+	transfer_unregister(transfer);
 }
 
 static DBusMessage *pull_phonebook(struct session_data *session,
@@ -456,7 +478,6 @@ static DBusMessage *pull_phonebook(struct session_data *session,
 				"Failed");
 
 	session->msg = dbus_message_ref(message);
-	session->filled = 0;
 
 	return NULL;
 }
@@ -522,7 +543,6 @@ static DBusMessage *pull_vcard_listing(struct session_data *session,
 				"Failed");
 
 	session->msg = dbus_message_ref(message);
-	session->filled = 0;
 
 	return NULL;
 }
@@ -671,7 +691,7 @@ static DBusMessage *pbap_select(DBusConnection *connection,
 	if (err)
 		return g_dbus_create_error(message,
 				ERROR_INF ".Failed",
-				OBEX_ResponseToString(err));
+				"%s", OBEX_ResponseToString(err));
 
 	return dbus_message_new_method_return(message);
 }
@@ -707,7 +727,8 @@ static DBusMessage *pbap_pull_vcard(DBusConnection *connection,
 
 	if (!pbapdata->path)
 		return g_dbus_create_error(message,
-				ERROR_INF ".Forbidden", "Call Select first of all");
+				ERROR_INF ".Forbidden",
+				"Call Select first of all");
 
 	if (dbus_message_get_args(message, NULL,
 			DBUS_TYPE_STRING, &name,
@@ -735,7 +756,6 @@ static DBusMessage *pbap_pull_vcard(DBusConnection *connection,
 				"Failed");
 
 	session->msg = dbus_message_ref(message);
-	session->filled = 0;
 
 	return NULL;
 }
